#include "miniev.h"
#include "asout.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <signal.h>
#include <errno.h>
#include <pthread.h>

#define ECHO_BUFFER_SIZE (1024*1024)

typedef struct echo_t {
	char buffer[ECHO_BUFFER_SIZE];
	asout_t* out;
} echo_t;

void* echo_create(asout_t* out) {
	echo_t* e = malloc(sizeof(echo_t));
	e->out = out;
	return e;
}

void echo_clean(void* e) {
	free(e);
}

bool echo_handler(int fd, void* arg) {
	echo_t* e = (echo_t*)arg;

	ssize_t bytes = read(fd, e->buffer, ECHO_BUFFER_SIZE);
	if(bytes <= 0) {
		if(bytes != 0 && (errno == EAGAIN || errno == EINTR)) {
			return true;
		}
		return false;
	}

	asout_send(e->out, fd, e->buffer, bytes);

	return true;
}



static void stop_server();

typedef struct accept_t {
	miniev_t* mev;
	asout_t* out;
} accept_t;

void* accept_create(miniev_t* mev, asout_t* out) {
	accept_t* a = malloc(sizeof(accept_t));
	a->mev = mev;
	a->out = out;
	return a;
}

void accept_clean(void* arg) {
	free(arg);
	stop_server();
}

bool accept_handler(int lfd, void* arg) {
	accept_t* a = (accept_t*)arg;

	int fd = accept(lfd, NULL, NULL);
	if(fd <= 0) {
		if(fd != 0 && (errno == EAGAIN || errno == EINTR)) {
			return true;
		}
		return false;
	}

	void* e = echo_create(a->out);
	miniev_cb_t cb = {e, &echo_handler, &echo_clean};
	miniev_push(a->mev, fd, cb);

	return true;
}



volatile sig_atomic_t g_stop_flag = false;

static void stop_server() {
	g_stop_flag = true;
}

void* event_thread(void* arg) {
	int lfd = *(int*)arg;

	miniev_t* mev = miniev_new();
	asout_t* out = asout_new();

	miniev_cb_t cbo = {out, &asout_handler, NULL};
	miniev_push(mev, asout_get_fd(out), cbo);

	void* a = accept_create(mev, out);
	miniev_cb_t cb = {a, &accept_handler, &accept_clean};
	miniev_push(mev, lfd, cb);

	while(g_stop_flag == false) {
		miniev_next(mev, 1000);
	}

	miniev_free(mev);
	asout_free(out);

	return NULL;
}



static void on_signal(int signo) {
	stop_server();
}

int main(int argc, char* argv[])
{
	int lfd = socket(PF_INET, SOCK_STREAM, 0);
	
	struct sockaddr_in addr_listen;
	addr_listen.sin_family = AF_INET;
	addr_listen.sin_addr.s_addr = INADDR_ANY;
	addr_listen.sin_port = htons(9090);

	int on = 1;
	setsockopt(lfd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

	bind(lfd, (struct sockaddr*)&addr_listen, sizeof(addr_listen));
	listen(lfd, 2048);

	signal(SIGINT,  &on_signal);
	signal(SIGTERM, &on_signal);
	signal(SIGPIPE, SIG_IGN);

	int num_thread = 4;
	pthread_t th[num_thread];

	for(int i = 0; i < num_thread-1; ++i) {
		pthread_create(&th[i], NULL, &event_thread, &lfd);
	}
	event_thread(&lfd);

	for(int i = 0; i < num_thread-1; ++i) {
		pthread_join(th[i], NULL);
	}

	return 0;
}

